import os, json, logging, requests
from socket import gaierror
from requests.auth import HTTPBasicAuth
    
def post(url, headers=None, data=None, rjson=None, raw=False, auth=None, v2=False):
    if rjson is not None:
        json.dumps(rjson)
    return call("POST", url, headers, data, rjson, raw, auth)

def patch(url, headers=None, data=None, rjson=None, raw=False, auth=None, v2=False):
    if rjson is not None:
        json.dumps(rjson)
    return call("PATCH", url, headers, data, rjson, raw, auth)

def get(url, headers=None, data=None, rjson=None, raw=False, auth=None, v2=False):
    if rjson is not None:
        json.dumps(rjson)
    return call("GET", url, headers, data, rjson, raw, auth)

def put(url, headers, data=None, rjson=None, raw=False, auth=None, v2=False):
    if rjson is not None:
        json.dumps(rjson)
    return call("PUT", url, headers, data, rjson, raw, auth)

def delete(url, headers, data=None, rjson=None, raw=False, auth=None, v2=False):
    if rjson is not None:
        json.dumps(rjson)
    return call("DELETE", url, headers, data, rjson, raw, auth)

def call(req_type, req_url, req_headers=None, req_data=None, req_json=None, raw=False, req_auth=None, v2=False):
    if not v2:
        if raw:
            try: 
                if type(req_auth) is dict:
                    response = requests.request(method=req_type, url=req_url, headers=req_headers, data=req_data, json=req_json, auth=HTTPBasicAuth(req_auth['user'], req_auth['pass'])) 
                else:
                    response = requests.request(method=req_type, url=req_url, headers=req_headers, data=req_data, json=req_json) 
                info(f"\n{response.json}")
                return response.json
            except:
                critical(f"Failure running API call - code: {response}")
                return None
        try:
            if type(req_auth) is dict:
                response = requests.request(method=req_type, url=req_url, headers=req_headers, data=req_data, json=req_json, auth=HTTPBasicAuth(req_auth['user'], req_auth['pass'])) 
            else:
                response = requests.request(method=req_type, url=req_url, headers=req_headers, data=req_data, json=req_json) 
        except gaierror:
            critical(f"Unable to resolve API endpoint - {req_url} - possible issue with DNS")
            return None
        except:
            critical(f"Failure running API call - code: {response}")
            return None
        try:
            return response.json()
        except:
            return response
    else:
        try:
            if type(req_auth) is dict:
                response = requests.request(method=req_type, url=req_url, headers=req_headers, data=req_data, json=req_json, auth=HTTPBasicAuth(req_auth['user'], req_auth['pass']))
            else:
                response = requests.request(method=req_type, url=req_url, headers=req_headers, data=req_data, json=req_json) 
            return response
        except gaierror:
            critical(f"Unable to resolve API endpoint - {req_url} - possible issue with DNS")
        except:
            critical(f"Failure running API call - code: {response}")     
            
#### logging

LOGNAME = 'curl.log'
DICTCONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s %(message)s'
        },
    },
    'handlers': {
        'file_handler': {
            'level': 'INFO',
            'mode': 'a',
            'filename': f"{os.getenv('HOME')}/.{LOGNAME}",
            'class': 'logging.FileHandler',
            'formatter': 'standard'
        }
    },
    'loggers': {
        '': {
            'handlers': ['file_handler'],
            'level': 'INFO',
            'propagate': True
        },
    }
}
logging.config.dictConfig(DICTCONFIG)

def error(msg):
    logging.error(msg)
    if logging.root.level == logging.INFO:
        print(f"ERROR: {msg}")

def info(msg, output=False):
    logging.info(msg)
    if logging.root.level == logging.INFO or output:
        print(f"INFO: {msg}")

def warn(msg):
    logging.warn(msg)
    if logging.root.level == logging.INFO:
        print(f"WARN: {msg}")

def debug(msg):
    logging.debug(msg)
    if logging.root.level == logging.DEBUG:
        print(f"DEBUG: {msg}")
    
def critical(msg):
    logging.critical(msg)
    logging.debug('exception: ', exc_info=True)
    print(f"CRITICAL: {msg}")



# these are some examples of click auto completion for projects and slack emoji names
import os, glob
from pathlib import Path
from configstore.configstore import Config
from toolbox.misc import detect_environment
try:
    import importlib_resources as resources
except:
    from importlib import resources

def complete_oci_profiles(ctx, param, incomplete):
    CONFIG = Config('ocitools')
    DATA = []
    NAMES = []
    for PROFILE in CONFIG.PROFILES:
        if 'latest' not in PROFILE:
            NAMES.append(PROFILE.strip())
    if NAMES:
        return [k for k in NAMES if k.startswith(incomplete)]
    else:
        return None

def complete_aws_profiles(ctx, param, incomplete):
    CONFIG = Config('awstools')
    DATA = []
    NAMES = []
    for PROFILE in CONFIG.PROFILES:
        if 'latest' not in PROFILE:
            NAMES.append(PROFILE.strip())
    if NAMES:
        return [k for k in NAMES if k.startswith(incomplete)]
    else:
        return None

def complete_aws_regions(ctx, param, incomplete):
    CONFIG = Config('awstools')
    DATA = []
    NAMES = []
    for PROFILE in CONFIG.PROFILES:
        if 'latest' not in PROFILE:
            NAMES.append(PROFILE.strip())
    for PROFILE in NAMES:
        try:
            for ENTRY in CONFIG.PROFILES[PROFILE]['metadata']['cached_regions']:
                DATA_ENTRY = {}
                if ENTRY != 'last_cache_update':
                    DATA.append(ENTRY)
        except:
            pass
    if DATA:
        return [k for k in DATA if k.startswith(incomplete)]
    else:
        return None

def complete_oci_regions(ctx, param, incomplete):
    CONFIG = Config('ocitools')
    DATA = []
    NAMES = []
    for PROFILE in CONFIG.PROFILES:
        if 'latest' not in PROFILE:
            NAMES.append(PROFILE.strip())
    for PROFILE in NAMES:
        try:
            for ENTRY in CONFIG.PROFILES[PROFILE]['metadata']['cached_regions']:
                DATA_ENTRY = {}
                if ENTRY != 'last_cache_update':
                    DATA.append(ENTRY)
        except:
            pass
    if DATA:
        return [k for k in DATA if k.startswith(incomplete)]
    else:
        return None

def complete_idp_profiles(ctx, param, incomplete):
    CONFIG = Config('idptools')
    NAMES = []
    for PROFILE in CONFIG.PROFILES:
        NAMES.append(PROFILE)
    return [k for k in NAMES if k.startswith(incomplete)]

def complete_csp_profiles(ctx, param, incomplete):
    CONFIG = Config('csptools')
    NAMES = []
    for PROFILE in CONFIG.PROFILES:
        NAMES.append(PROFILE)
    return [k for k in NAMES if k.startswith(incomplete)]

def complete_jenkins_profiles(ctx, param, incomplete):
    CONFIG = Config('jenkinstools')
    NAMES = []
    for PROFILE in CONFIG.PROFILES:
        NAMES.append(PROFILE)
    return [k for k in NAMES if k.startswith(incomplete)]

def complete_jira_profiles(ctx, param, incomplete):
    CONFIG = Config('jiratools')
    CACHED_PROFILES = []
    for PROFILE in CONFIG.PROFILES:
        CACHED_PROFILES.append(PROFILE)
    return [k for k in CACHED_PROFILES if k.startswith(incomplete)]

def complete_configstore_names(ctx, param, incomplete):
    NAMES = []
    HOME = str(Path.home())
    for KEY in glob.glob(HOME + "/goat/.*key"):
        if os.path.isfile(KEY):
            KEY = str(Path(KEY).stem.replace(".",""))
            BASE = os.path.basename(KEY)
            NAMES.append(str(BASE))
    return [k for k in NAMES if k.startswith(incomplete)]

def complete_projects(ctx, param, incomplete):
    from jiratools.auth import get_default_profile
    CONFIG = Config('jiratools')
    CACHED_PROJECTS = {}
    CACHED_PROJECTS.update(CONFIG.get_metadata('projects', get_default_profile()))
    return [k for k in CACHED_PROJECTS if k.startswith(incomplete)]

def complete_emojis(ctx, param, incomplete):
    EMOJIS = {}
    EMOJIS = listAllEmojis()
    return [k for k in EMOJIS if k.startswith(incomplete)]

def listAllEmojis():
    EMOJIS = []
    MY_RESOURCES = resources.files("toolbox")
    DATA = (MY_RESOURCES / "emoji.lst")
    with open(DATA) as f:
        CONTENTS = f.read()
        for line in CONTENTS.split('\n'):
            if line:
                EMOJIS.append(line)
    return EMOJIS



import sys, os, signal, readline

if os.name != 'nt':
    import tty

from prompt_toolkit import prompt
from prompt_toolkit.filters import Condition
from prompt_toolkit.keys import Keys
from prompt_toolkit.key_binding import KeyBindings

user_env_var = "USERNAME" if os.name == 'nt' else "LOGNAME"
default_assignee = os.environ.get(user_env_var, None)

def getJiraCredentials():
    try:
        JIRA_USER = default_assignee
        JIRA_PASS = os.environ['JIRA_API_TOKEN']
    except:
        JIRA_USER = ''
        JIRA_PASS = ''
    return JIRA_USER, JIRA_PASS

def getCreds():

    JIRA_USER, JIRA_PASS = getJiraCredentials()
    if JIRA_USER and JIRA_PASS:
        return JIRA_USER, JIRA_PASS
    hidden = [True]  # Nonlocal
    bindings = KeyBindings()

    if os.name != 'nt':
        signal.signal(signal.SIGTSTP, signal.SIG_IGN)
    stdin = sys.__stdin__.fileno()
    stream = sys.__stderr__.fileno()

    if os.name != 'nt':
        old = tty.tcgetattr(stdin)

    JIRA_USER = input('Enter username ' + "[" + str(default_assignee).split('@')[0] + "] : ").strip()
    if not JIRA_USER:
        JIRA_USER = str(default_assignee).split('@')[0]

    @bindings.add("c-t")
    def _(event):
        "When ControlT has been pressed, toggle visibility."
        hidden[0] = not hidden[0]

    JIRA_PASS = prompt(
        "Enter password: ", is_password=Condition(lambda: hidden[0]), key_bindings=bindings
    )
    # restore terminal settings
    if os.name != 'nt':
        tty.tcsetattr(stdin, tty.TCSAFLUSH, old)
    # enable (^Z) SIGTSTP
    if os.name != 'nt':
        signal.signal(signal.SIGTSTP, signal.SIG_DFL)

    # return credentials
    return JIRA_USER, JIRA_PASS

# generic version of the two functions above
def get_secure_string_from_environ(var_name):
    try:
        SECURE_STRING = os.environ[var_name]
    except:
        SECURE_STRING = None
    return SECURE_STRING

def get_secure_string(var_name, prompt_msg="Enter password: "):
    SECURE_STRING = get_secure_string_from_environ(var_name)
    if SECURE_STRING:
        return SECURE_STRING
    HIDDEN = [True]
    BINDINGS = KeyBindings()
    signal.signal(signal.SIGTSTP, signal.SIG_IGN)
    STDIN = sys.__stdin__.fileno()
    STREAM = sys.__stderr__.fileno()
    if os.name != 'nt':
        OLD = tty.tcgetattr(STDIN)
    @BINDINGS.add("c-t")
    def _(event):
        "When ControlT has been pressed, toggle visibility."
        HIDDEN[0] = not HIDDEN[0]
    SECURE_STRING = prompt(
        prompt_msg, is_password=Condition(lambda: HIDDEN[0]), key_bindings=BINDINGS
    )
    if os.name != 'nt':
        tty.tcsetattr(STDIN, tty.TCSAFLUSH, OLD)
    signal.signal(signal.SIGTSTP, signal.SIG_DFL)
    return SECURE_STRING

def password_prompt():

    hidden = [True]  # Nonlocal
    bindings = KeyBindings()

    signal.signal(signal.SIGTSTP, signal.SIG_IGN)
    stdin = sys.__stdin__.fileno()
    stream = sys.__stderr__.fileno()

    if os.name != 'nt':
        old = tty.tcgetattr(stdin)

    @bindings.add("c-t")
    def _(event):
        "When ControlT has been pressed, toggle visibility."
        hidden[0] = not hidden[0]

    PASSWORD = prompt(
        "Enter password: ", is_password=Condition(lambda: hidden[0]), key_bindings=bindings
    )
    # restore terminal settings
    if os.name != 'nt':
        tty.tcsetattr(stdin, tty.TCSAFLUSH, old)
        # enable (^Z) SIGTSTP
        signal.signal(signal.SIGTSTP, signal.SIG_DFL)

    # return credentials
    return PASSWORD

def getOtherCreds(title='default'):

    hidden = [True]  # Nonlocal
    bindings = KeyBindings()

    signal.signal(signal.SIGTSTP, signal.SIG_IGN)
    stdin = sys.__stdin__.fileno()
    stream = sys.__stderr__.fileno()

    if os.name != 'nt':
        old = tty.tcgetattr(stdin)

    USER = input('Enter ' + title + ' username ' + "[" + str(default_assignee).split('@')[0] + "] : ").strip()
    if not USER:
        USER = str(default_assignee).split('@')[0]

    @bindings.add("c-t")
    def _(event):
        "When ControlT has been pressed, toggle visibility."
        hidden[0] = not hidden[0]

    PASS = prompt(
        "Enter " + title + " password: ", is_password=Condition(lambda: hidden[0]), key_bindings=bindings
    )
    if os.name != 'nt':
        # restore terminal settings
        tty.tcsetattr(stdin, tty.TCSAFLUSH, old)
        # enable (^Z) SIGTSTP
        signal.signal(signal.SIGTSTP, signal.SIG_DFL)

    # return credentials
    return USER, PASS

def getFullUrl(title='default'):

    hidden = [True]  # Nonlocal
    bindings = KeyBindings()

    signal.signal(signal.SIGTSTP, signal.SIG_IGN)
    stdin = sys.__stdin__.fileno()
    stream = sys.__stderr__.fileno()

    if os.name != 'nt':
        old = tty.tcgetattr(stdin)

    @bindings.add("c-t")
    def _(event):
        "When ControlT has been pressed, toggle visibility."
        hidden[0] = not hidden[0]

    URL = prompt(
        "Paste " + title + " full URL here: ", is_password=Condition(lambda: hidden[0]), key_bindings=bindings
    )

    if os.name != 'nt':
        # restore terminal settings
        tty.tcsetattr(stdin, tty.TCSAFLUSH, old)
        # enable (^Z) SIGTSTP
        signal.signal(signal.SIGTSTP, signal.SIG_DFL)

    # return URL
    return remove_lead_and_trail_slash(URL)

def remove_lead_and_trail_slash(s):
    if s.startswith('/'):
        s = s[1:]
    if s.endswith('/'):
        s = s[:-1]
    return s

def getOtherToken(title='default'):

    hidden = [True]  # Nonlocal
    bindings = KeyBindings()

    signal.signal(signal.SIGTSTP, signal.SIG_IGN)
    stdin = sys.__stdin__.fileno()
    stream = sys.__stderr__.fileno()

    if os.name != 'nt':
        old = tty.tcgetattr(stdin)

    @bindings.add("c-t")
    def _(event):
        "When ControlT has been pressed, toggle visibility."
        hidden[0] = not hidden[0]

    TOKEN = prompt(
        "Paste " + title + " token here: ", is_password=Condition(lambda: hidden[0]), key_bindings=bindings
    )
    if os.name != 'nt':
        # restore terminal settings
        tty.tcsetattr(stdin, tty.TCSAFLUSH, old)
        # enable (^Z) SIGTSTP
        signal.signal(signal.SIGTSTP, signal.SIG_DFL)

    # return credentials
    return TOKEN



import os, sys
if os.name != 'nt':
    try:
       import curses
    except:
        pass
from math import *
from tabulate import tabulate

os.environ['NCURSES_NO_UTF8_ACS'] = '1'

class Menu():

    def __init__(self, options, title, joiner, subtitle):
        self.ATTRIBUTES = {}
        self.OPTIONS = options
        self.TITLE = title
        self.JOINER = joiner # this specifies how to join your strings; i.e. '\t\t'
        self.SUBTITLE = subtitle
        
    def _setup_menu(self, stdscr):

        rows, columns = os.popen('stty size', 'r').read().split()
        screen = curses.initscr()
        screen.clrtoeol()
        screen.erase()
        curses.noecho()
        curses.cbreak()
        curses.start_color()
        screen.keypad( 1 )
        highlightText = curses.color_pair( 1 )
        normalText = curses.A_NORMAL
        screen.border(0)
        height, width = stdscr.getmaxyx()

        # title information
        title = self.TITLE
        author = "author: GOAT Team"
        subtitle = self.SUBTITLE
      
        max_row = 15 # max number of rows
        box = curses.newwin( max_row + 2, int(columns) - 2, 2, 1 )
        box.box()

        # decorate the titles and enable color
        curses.curs_set(0)
        curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_CYAN)
        curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_RED)
        curses.init_pair(3, curses.COLOR_BLACK, curses.COLOR_WHITE)
        curses.init_pair(4, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.curs_set(0)
        stdscr.addstr(1, 1, title, curses.color_pair(4))
        stdscr.addstr(1, width - len(author) - 2, author, curses.color_pair(4))
        stdscr.addstr(1, int(columns) // 2 - len(subtitle) // 2, subtitle, curses.color_pair(4))

        x = 0
        cursor_x = 0
        cursor_y = 0

        strings = self.OPTIONS
        joinit = self.JOINER
        if not joinit:
            joinit = '\t'
        row_num = len( strings )
        pages = int( ceil( row_num / max_row ) )
        position = 1
        page = 1

        for i in range( 1, max_row + 1 ):
            if row_num == 0:
                box.addstr( 1, 1, "THESE AREN'T STRINGS - PRESS 'Q' TO QUIT", highlightText )
            else:
                size = len(strings[ i - 1]) + 5
                space = " " * size
                if (i == position):
                    box.addstr( i, 2, str( i ) + " - " + f'{joinit}'.join(strings[ i - 1 ]), highlightText)
                else:
                    box.addstr( i, 2, str( i ) + " - " + f'{joinit}'.join(strings[ i - 1 ]), normalText)
                if i == row_num:
                    break

        # Render status bar with subtitle
        statusbarstr = 'PRESS "q" TO QUIT'
        stdscr.attron(curses.color_pair(3))
        height = int(height) - 1
        stdscr.addstr(height-1, 1, statusbarstr)
        stdscr.attroff(curses.color_pair(3))

        screen.refresh()
        box.refresh()
        x = screen.getch()

        while x != 27 and x != ord('q'):
            screen.refresh()
            box.refresh()
            height, width = stdscr.getmaxyx()
            if x == curses.KEY_DOWN:
                cursor_y = cursor_y + 1
                if page == 1:
                    if position < i:
                        position = position + 1
                    else:
                        if pages > 1:
                            page = page + 1
                            position = 1 + ( max_row * ( page - 1 ) )
                elif page == pages:
                    if position < row_num:
                        position = position + 1
                else:
                    if position < max_row + ( max_row * ( page - 1 ) ):
                        position = position + 1
                    else:
                        page = page + 1
                        position = 1 + ( max_row * ( page - 1 ) )
            if x == curses.KEY_UP:
                cursor_y = cursor_y - 1
                if page == 1:
                    if position > 1:
                        position = position - 1
                else:
                    if position > ( 1 + ( max_row * ( page - 1 ) ) ):
                        position = position - 1
                    else:
                        page = page - 1
                        position = max_row + ( max_row * ( page - 1 ) )
            if x == curses.KEY_LEFT or x == curses.KEY_PPAGE:
                cursor_x = cursor_x - 1
                if page > 1:
                    page = page - 1
                    position = 1 + ( max_row * ( page - 1 ) )

            if x == curses.KEY_RIGHT or x == curses.KEY_NPAGE:
                cursor_x = cursor_x + 1
                if page < pages:
                    page = page + 1
                    position = ( 1 + ( max_row * ( page - 1 ) ) )
            if x == ord( "\n" ) and row_num != 0:
                screen.erase()
                return strings[ position - 1]

            box.erase()
            screen.border( 0 )
            box.border( 0 )
        
            for i in range( 1 + ( max_row * ( page - 1 ) ), max_row + 1 + ( max_row * ( page - 1 ) ) ):
                if row_num == 0:
                    box.addstr( 1, 1, "data must be strings",  highlightText )
                else:
                    if ( i + ( max_row * ( page - 1 ) ) == position + ( max_row * ( page - 1 ) ) ):
                        box.addstr( i - ( max_row * ( page - 1 ) ), 2, str( i ) + " - " + f'{joinit}'.join(strings[ i - 1 ]), highlightText )
                    else:
                        box.addstr( i - ( max_row * ( page - 1 ) ), 2, str( i ) + " - " + f'{joinit}'.join(strings[ i - 1 ]), normalText )
                    if i == row_num:
                        break

            # put the title back on refresh
            stdscr.addstr(1, 1, title, curses.color_pair(4))
            stdscr.addstr(2, 1, author, curses.color_pair(4))

            # Render status bar
            statusbarstr = 'PRESS "q" TO QUIT'
            stdscr.attron(curses.color_pair(3))
            height = int(height) - 1
            stdscr.addstr(height-1, 1, statusbarstr)
            stdscr.attroff(curses.color_pair(3))

            # Refresh the screen
            screen.refresh()
            box.refresh()
            x = screen.getch()

    def display(self):
        CHOICE = curses.wrapper(self._setup_menu)
        if CHOICE == "ERROR_DISPLAY":
            print("Oops, your window is too small to display the menu. Please maximize your terminal/client app and try again")
            sys.exit()
        curses.endwin()
        return CHOICE



import os, sys
import logging, logging.config

# reduce a json, aka keep only specified key:value pairs and discard others
# definitino dict shall be a collection of keys with either a none or a list value
# each key should hold a value None unless you intend to re-map a key from a source json
# for example, lets assume we have a json: { name: test, data: { config: { created: today } } }
# to produce a "reduced" json like so: { name: test, data: today } you'd need to pass a definition dict as follows:
# { name: None, data: ( data,config,created ) }
# lastly, json_dict can be set to True to reduce a "json of jsons"; ie { John: { age: 15, ht: 180 }, Sam: { age: 20, ht: 200 } }
def reduce_json(source_data, definition_dict, json_dict=False):
    if type(source_data) is list:
        RESULT = []
        for DATA in source_data:
            REDUCED_DATA = _reduce_json(DATA, definition_dict)
            RESULT.append(REDUCED_DATA)
    elif json_dict:
        RESULT = {}
        for KEY in source_data:
            REDUCED_DATA = _reduce_json(source_data[KEY], definition_dict)
            RESULT[KEY] = REDUCED_DATA
    else:
        RESULT = _reduce_json(source_data, definition_dict)
    return RESULT

# worker function for reduce_json
def _reduce_json(source_json, definition_dict):
    REDUCED_DATA = {}
    for KEY in definition_dict:
        VALUE = definition_dict[KEY]
        if VALUE is None:
            if KEY in source_json:
                REDUCED_DATA[KEY] = source_json[KEY]
            else:
                REDUCED_DATA[KEY] = ""
        elif type(VALUE) == dict:
            SUB_DATA = source_json[KEY]
            REDUCED_SUB_DATA = _reduce_json(SUB_DATA, VALUE)
            REDUCED_DATA[KEY] = REDUCED_SUB_DATA
        elif type(VALUE) is list:
            SUB_DATA = source_json
            for ENTRY in VALUE:
                try:
                    if ENTRY in SUB_DATA:
                        SUB_DATA = SUB_DATA[ENTRY]
                    else:
                        REDUCED_DATA[KEY] = None
                        break
                except TypeError:
                    REDUCED_DATA[KEY] = ""
            REDUCED_DATA[KEY] = SUB_DATA
        else:
            critical('error parsing json - toolbox.jsontools.filter_json')
    return REDUCED_DATA

def filter(source_data, filter, json_dict=False):
    if type(source_data) is list:
        RESULT = []
        for DATA in source_data:
            REDUCED_DATA = filter_json(DATA, filter)
            if REDUCED_DATA is not None:
                RESULT.append(REDUCED_DATA)
    elif json_dict:
        RESULT = {}
        for KEY in source_data:
            REDUCED_DATA = filter_json(source_data[KEY], filter)
            if REDUCED_DATA is not None:
                RESULT[KEY] = REDUCED_DATA
    else:
        RESULT = filter_json(source_data, filter)
    return RESULT

def filter_json(data, filter):
    FILTER_RESULTS = []
    if ',' in filter:
        FILTER = filter.split(',')
    else:
        FILTER = [ filter ] 
    for ITEM in FILTER:
        if ':' not in ITEM:
            warn(f'detected incorrect filter item {ITEM}; ignoring it')
            warn('to filter the json for key-value pairs, use "key:pair" format')
            warn('to look just for a key, or just for a value use: "key:" or ":value"' )
            break
        KEY,VALUE = ITEM.split(':',1)
        if ':' in VALUE:
            warn(f'detected incorrect filter item {ITEM}; ignoring it')
            warn('it is not allowed to use ":" character more than once')
            break
        FILTER_RESULTS.append(_filter_json(data, KEY, VALUE))
    for RESULT in FILTER_RESULTS:
        if RESULT == False:
            return None
    return data
        
def _filter_json(data, filter_key, filter_value):
    if filter_key != "" and filter_value != "":
        FOUND_KEY = False
        FOUND_VALUE = False
        for KEY in data:
            if KEY == filter_key:
                FOUND_KEY = True
            VALUE = data[KEY]
            if VALUE == filter_value:
                FOUND_VALUE = True
            elif type(VALUE) is dict:
                _filter_json(VALUE, filter_key, filter_value)
            else:
                FOUND_KEY = False
                FOUND_VALUE = False
            if FOUND_KEY and FOUND_VALUE:
                return True
        return False
    elif filter_key == "" and filter_value != "":
        FOUND_VALUE = False
        for KEY in data:
            VALUE = data[KEY]
            if VALUE == filter_value:
                FOUND_VALUE = True
            elif type(VALUE) is dict:
                _filter_json(VALUE, filter_key, filter_value)
            else:
                FOUND_VALUE = False
            if FOUND_VALUE:
                return True
        return False
    elif filter_key != "" and filter_value == "":
        FOUND_KEY = False
        for KEY in data:
            VALUE = data[KEY]
            if KEY == filter_key:
                FOUND_KEY = True
            elif type(VALUE) is dict:
                _filter_json(VALUE, filter_key, filter_value)
            else:
                FOUND_KEY = False
            if FOUND_KEY:
                return True
        return False
    else:
        return False

#### logging

LOGNAME = 'jsonfilter.log'
DICTCONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s %(message)s'
        },
    },
    'handlers': {
        'file_handler': {
            'level': 'INFO',
            'mode': 'a',
            'filename': f"{os.getenv('HOME')}/goat/.{LOGNAME}",
            'class': 'logging.FileHandler',
            'formatter': 'standard'
        }
    },
    'loggers': {
        '': {
            'handlers': ['file_handler'],
            'level': 'INFO',
            'propagate': True
        },
    }
}
logging.config.dictConfig(DICTCONFIG)

def error(msg):
    logging.error(msg)
    if logging.root.level == logging.INFO:
        print(f"ERROR: {msg}")

def info(msg, output=False):
    logging.info(msg)
    if logging.root.level == logging.INFO or output:
        print(f"INFO: {msg}")

def warn(msg):
    logging.warn(msg)
    if logging.root.level == logging.INFO:
        print(f"WARN: {msg}")

def debug(msg):
    logging.debug(msg)
    if logging.root.level == logging.DEBUG:
        print(f"DEBUG: {msg}")
    
def critical(msg):
    logging.critical(msg)
    logging.debug('exception: ', exc_info=True)
    print(f"CRITICAL: {msg}")
    sys.exit()



from toolbox.logger import Log

class nv():
    # name-value pairs
    def __init__(self, *args):
        for arg in args:
            if type(arg) is list and len(arg) is 2:
                NAME = arg[0]
                VALUE = arg[1]
                setattr(self, NAME, VALUE)

class rv():
    # binary result (true/false) and a value (message, code etc)
    def __init__(self, result, value):
        if type(result) is bool:
            self.ok = result
            self.msg = value
    def print(self, quit_on_fail=True):
        if self.ok:
            Log.info(self.msg)
        elif quit_on_fail:
            Log.critical(self.msg)
        else:
            Log.warn(self.msg)

class rnv():
    # binary result and name-vlaue pairs
    def __init___(self, result, *args):
        if type(result) is bool:
            self.ok = result
            for arg in args:
                if type(arg) is list and len(arg) is 2:
                    NAME = arg[0]
                    VALUE = arg[1]
                    setattr(self, NAME, VALUE)
    def print(self, quit_on_fail=True):
        if self.ok:
            Log.info(self.msg)
        elif quit_on_fail:
            Log.critical(self.msg)
        else:
            Log.warn(self.msg)



import os
import sys
import platform
import logging, logging.config
import json

class Log:
    def __init__(self, logname, level='1'):
        self.NAME = logname
        MODES = {
            '0': 'WARN',
            '1': 'INFO',
            '2': 'DEBUG'
        }
        self.MODE = MODES[str(level)]

    # V2 
    def setup(name,level=1):
        MODES = {
            0: 'WARN',
            1: 'INFO',
            2: 'DEBUG'
        }
        MODE = MODES[level]
        try:
            DICTCONFIG = {
                'version': 1,
                'disable_existing_loggers': False,
                'formatters': {
                    'standard': {
                        'format': '%(asctime)s %(levelname)s %(message)s'
                    },
                },
                'handlers': {
                    'file_handler': {
                        'level': MODE,
                        'mode': 'a',
                        'filename': f"{os.getenv('HOME')}/goat/.{name}.log",
                        'class': 'logging.FileHandler',
                        'formatter': 'standard'
                    }
                },
                'loggers': {
                    '': {
                        'handlers': ['file_handler'],
                        'level': MODE,
                        'propagate': True
                    },
                }
            }
            logging.config.dictConfig(DICTCONFIG)
        except FileExistsError:
            sys.exit("ERROR: Unable to setup a logfile. Exiting")

    def startlog(self):    
        try:
            DICTCONFIG = {
                'version': 1,
                'disable_existing_loggers': False,
                'formatters': {
                    'standard': {
                        'format': '%(asctime)s %(levelname)s %(message)s'
                    },
                },
                'handlers': {
                    'file_handler': {
                        'level': self.MODE,
                        'mode': 'a',
                        'filename': f"{os.getenv('HOME')}/goat/.{self.NAME}.log",
                        'class': 'logging.FileHandler',
                        'formatter': 'standard'
                    }
                },
                'loggers': {
                    '': {
                        'handlers': ['file_handler'],
                        'level': self.MODE,
                        'propagate': True
                    },
                }
            }
            logging.config.dictConfig(DICTCONFIG)
        except FileExistsError:
            sys.exit("ERROR: Unable to setup a logfile. Exiting")

    def error(msg):
        logging.error(msg)
        if logging.root.level == logging.INFO:
            print(f"ERROR: {msg}")

    def json(msg, output=False):
        logging.info(msg)
        if logging.root.level == logging.INFO or output:
            print(f"{msg}")

    def info(msg, output=False):
        logging.info(msg)
        if logging.root.level == logging.INFO or output:
            print(f"INFO: {msg}")

    def warn(msg):
        logging.warn(msg)
        if logging.root.level == logging.INFO:
            print(f"WARN: {msg}")

    def debug(*args):
        for arg in args:
            logging.debug(arg)
            if logging.root.level == logging.DEBUG:
                print(f"DEBUG: {arg}")
        
    def notify(msg, title, subtitle, link, cmd):
        # for mac raise notification for critical issues
        RES = platform.system()
        if 'Darwin' in RES:
            from pync import Notifier
            if link:
                try:
                    Notifier.notify(msg, title=title, subtitle=subtitle, open=link)
                except:
                    pass
            elif cmd:
                try:
                    Notifier.notify(msg, title=title, subtitle=subtitle, execute=cmd)
                except:
                    pass
            else:
                try:
                    Notifier.notify(msg, title=title, subtitle=subtitle)
                except:
                    pass

    def critical(msg):
        logging.critical(msg)
        logging.debug('exception: ', exc_info=True)
        print(f"CRITICAL: {msg}")
        sys.exit()
        



# misc functions
import sys, os, shutil, getpass, glob, gnupg, json, re, operator, base64
from toolbox.logger import Log
from toolbox.menumaker import Menu
from pathlib import Path

try:
    import importlib_resources as resources
except:
    from importlib import resources

MOVE = '\033[35G'
MOVE2 = '\033[75G'
HOME = os.getenv("HOME")
RESET = '\033[0m'
GREEN = '\033[0;32m'
RED = '\033[1;31m'
UNDERLINE = '\033[4m'
MYBLUE = '\033[0;2;60m'
debug = 1
pushstack = list()
SCREEN_WIDTH=80
centered = operator.methodcaller('center', SCREEN_WIDTH)

def draw_title():
    b = centered('''
        (_(
        /_/'_____/)
        "  |      |
           |""""""|
    ''')
    return(''.join(b))

def convert(list):
    # convert list to tuple
    return tuple(list)

def convertTuple(tup):
    # initialize an empty string
    str = ''
    for item in tup:
        str = str + item
    return str

def decode_string(string):
    string_bytes = string.encode('ascii')
    return base64.b64decode(string_bytes).decode('ascii')

def set_terminal_width():
    return max(90, shutil.get_terminal_size().columns - 2)

def get_aliases():
    COMPLETION = HOME + '/goat_completion.sh'
    if os.path.isfile(COMPLETION):
        with open(COMPLETION) as F:
            LINES = [LINE.rstrip() for LINE in F]
        ALIASES = []
        TOTAL = 0
        for L in LINES:
            if 'alias' in L:
                TOTAL = TOTAL + 1
                NAME = L.split('=')[0].replace('alias','').strip()
                VALUE = L.split('=')[1].replace('"','').replace('"','')
                ALIAS_DICT = {}
                ALIAS_DICT['alias'] = NAME
                ALIAS_DICT['command'] = VALUE
                ALIASES.append(ALIAS_DICT)
        if ALIASES:
            return ALIASES
        else:
            Log.info("there are no aliases defined; configuring them now")
            define(COMPLETION)
    else:
        Log.info("there are no aliases defined; configuring them now")
        define(COMPLETION)

def define(completion):
    OSNAME = getpass.getuser().split('@')[0]
    with open(completion, "at") as source:
        source.write(
            "\n"
            "# Added by goat #\n"
        )

def detect_environment():
    return 'non-gc'

def search_man_pages(manuals):

    NAMES = []
    PATTERNS = []
    MINE = []
    SAVE = []

    MY_RESOURCES = resources.files("goat")
    DATA = (MY_RESOURCES / "manuals")

    # compile a list of patterns
    for SEARCH in manuals:
        PATTERNS.append(SEARCH)

    PATTERNS.sort()

    # compile a list of man pages
    for MAN in os.listdir(DATA):
        MINE.append(MAN)

    RUN = {}
    for MAN in MINE:
        for PATTERN in PATTERNS:
            if MAN not in RUN and PATTERN in MAN:
                RUN[MAN] = {
                    "%s" %(PATTERN)
                }
            elif MAN in RUN and PATTERN in MAN:
                RUN[MAN].update({ "%s" %(PATTERN) })

    for MAN in RUN:
        p = RUN[MAN]
        p = [(v) for v in p]
        p.sort()
        if p == PATTERNS:
            SAVE.append(MAN)

    SAVE.sort()
    NAMES = SAVE

    if NAMES == []:
        Log.critical(f'WARN: unable to find any user manuals loaded matching pattern {manuals}')
    else:
        INPUT = 'viewer'
        CHOICE = runMenu(NAMES, INPUT)
        try:
            CHOICE = ''.join(CHOICE)
            name = CHOICE.split('\t')[0]
            if name:
                Log.info(f"gathering manual {name} now, please wait...")
            else:
                Log.critical("please select a user manual to continue...")
        except:
            Log.critical("please select a user manual to continue...")
    return (MY_RESOURCES / "manuals" / name)

def runMenu(DATA, INPUT):
    COUNT = 0
    FINAL = []
    TITLE = f'GOAT: {INPUT}'
    for data in DATA:
        COUNT = COUNT + 1
        RESULTS = []
        RESULTS.append(data)
        FINAL.append(RESULTS)
    SUBTITLE = f'showing {COUNT} available object(s)'
    JOINER = '\t\t'
    FINAL_MENU = Menu(FINAL, TITLE, JOINER, SUBTITLE)
    CHOICE = FINAL_MENU.display()
    return CHOICE

def remove_html_tags(text):
    import re
    text = re.sub('<.*?>', '', text)
    clean = re.sub('{panel.*?}', '', text)
    return clean.strip()

def pushd(dirname):
    global pushstack
    pushstack.append(os.getcwd())
    os.chdir(dirname)

def popd():
    global pushstack
    os.chdir(pushstack.pop())

def get_save_path(filename="aws.json"):
    user_home = Path.home()
    goat_shell_data_path = user_home / "goat" / "shell" / "data"
    goat_shell_data_path.mkdir(parents=True, exist_ok=True) # create directory if doesn't exist
    return goat_shell_data_path / filename

def attempt_load_custom_json(provider):
    """
    Attempt to load user's custom JSON from home directory.

    Args:
    - provider (str): The cloud provider's name.

    Returns:
    - bool: True if successful, False otherwise.
    """
    # Get the user's custom JSON path
    user_home = Path.home()
    user_json_path = user_home / "goat" / "shell" / "data" / f"{provider}.json"

    # If user's custom JSON does not exist, return False
    if not user_json_path.exists():
        return False

    # Try to load the JSON
    try:
        with user_json_path.open() as json_file:
            data = json.load(json_file)
        # Successfully loaded JSON
        return True
    except Exception as ex:
        logger.warning(f"Exception while attempting to load user's custom JSON for {provider}: {ex}")
        # Failed to load JSON
        return False

def is_valid_json(file_path):
    try:
        with open(file_path, 'r') as json_file:
            json.load(json_file)
        return True
    except:
        return False

def get_corrupt_provider_files():
    user_home = Path.home()
    data_directory = user_home / "goat" / "shell" / "data"
    corrupt_files_dict = {}

    for json_file in data_directory.glob('*.json'):
        if not is_valid_json(json_file):
            # Extracting the provider name from the file name
            provider_name = json_file.stem
            corrupt_files_dict[provider_name] = str(json_file)

    return corrupt_files_dict




#===========================================================================
#
# Extend Argparse to Enable Sub-Parser Groups
#
# Based on this very old issue: https://bugs.python.org/issue9341
#
# Adds the method `add_parser_group()` to the sub-parser class.
# This adds a group heading to the sub-parser list, just like the
# `add_argument_group()` method.
#
# NOTE: As noted on the issue page, this probably won't work with [parents].
# see http://bugs.python.org/issue16807
#
#===========================================================================
# Pylint doesn't like us access protected items like this
#pylint:disable=protected-access,abstract-method
import argparse
from argparse import HelpFormatter

class MyFormatter(argparse.HelpFormatter):
    def add_argument(self, action):
        if action.help is not argparse.SUPPRESS:

            # find all invocations
            get_invocation = self._format_action_invocation
            invocations = [get_invocation(action)]
            current_indent = self._current_indent
            for subaction in self._iter_indented_subactions(action):
                # compensate for the indent that will be added
                indent_chg = self._current_indent - current_indent
                added_indent = 'x'*indent_chg
                invocations.append(added_indent+get_invocation(subaction))
            # print('inv', invocations)# update the maximum item length
            invocation_length = max([len(s) for s in invocations])
            action_length = invocation_length + self._current_indent
            self._action_max_length = max(self._action_max_length,
                                          action_length)

            # add the item to the list
            self._add_item(self._format_action, [action])


class _SubParsersAction(argparse._SubParsersAction):
    class _PseudoGroup(argparse.Action):
        def __init__(self, container, title):
            sup = super(_SubParsersAction._PseudoGroup, self)
            sup.__init__(option_strings=[], dest=title)
            self.container = container
            self._choices_actions = []
        def add_parser(self, name, **kwargs):
            # add the parser to the main Action, but move the pseudo action
            # in the group's own list
            parser = self.container.add_parser(name, **kwargs)
            choice_action = self.container._choices_actions.pop()
            self._choices_actions.append(choice_action)
            return parser
        def _get_subactions(self):
            return self._choices_actions
        def add_parser_group(self, title):
            # the formatter can handle recursive subgroups
            grp = _SubParsersAction._PseudoGroup(self, title)
            self._choices_actions.append(grp)
            return grp
    def add_parser_group(self, title):
        #
        grp = _SubParsersAction._PseudoGroup(self, title)
        self._choices_actions.append(grp)
        return grp

class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register('action', 'parsers', _SubParsersAction)




import os, sys, logging, logging.config
from cryptography.fernet import Fernet

LOGNAME = 'fernet.log'
DICTCONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s %(message)s'
        },
    },
    'handlers': {
        'file_handler': {
            'level': 'INFO',
            'mode': 'a',
            'filename': f"{os.getenv('HOME')}/goat/.{LOGNAME}",
            'class': 'logging.FileHandler',
            'formatter': 'standard'
        }
    },
    'loggers': {
        '': {
            'handlers': ['file_handler'],
            'level': 'INFO',
            'propagate': True
        },
    }
}

HOME = os.getenv('HOME')
os.makedirs(f'{HOME}/goat', exist_ok=True)
logging.config.dictConfig(DICTCONFIG)

def setup(KEYNAME):
    KEYFILE = f"{os.getenv('HOME')}/goat/.{KEYNAME}.key" 
    if not os.path.exists(KEYFILE):
        warn("Encryption key not detected. Generating a new one")
        KEY = Fernet.generate_key()
        with open(KEYFILE, 'wb') as KEYFILE:
            try:
                KEYFILE.write(KEY)
            except:
                critical(f"Failed to generate encryption key")
    else:
        with open(KEYFILE, 'rb') as KEYFILE:
            try:
                KEY = KEYFILE.read()
            except:
                critical(f"Failed to read encryption key")
    try:
        CIPHER = Fernet(KEY)
    except:
        critical(f"Failed to start encryption system")
    return CIPHER

def encrypt_string(KEYNAME, INPUTSTR, FILEPATH=None):
    CIPHER = setup(KEYNAME)
    ENCRYPTED_STR = CIPHER.encrypt(bytes(INPUTSTR, 'utf-8'))
    if FILEPATH is None:
        return ENCRYPTED_STR.decode('utf-8')
    with open(FILEPATH, 'wb') as DESTINATION:
        DESTINATION.write(ENCRYPTED_STR)

def encrypt_file(INPUTPATH, OUTPUTPATH=None):
    with open(INPUTPATH, 'r+') as INPUTFILE:
        DECRYPTED_STR = INPUTFILE.read()
        ENCRYPTED_STR = encrypt_string(DECRYPTED_STR)
        if OUTPUTPATH is not None:
            with open(OUTPUTPATH, 'wb') as DESTINATION:
                DESTINATION.write(ENCRYPTED_STR)
        else:
            INPUTFILE.truncate(0)
            INPUTFILE.seek(0)
            INPUTFILE.write(ENCRYPTED_STR)

def decrypt_string(KEYNAME, INPUTSTR):
    CIPHER = setup(KEYNAME)
    DECRYPTED_STR = CIPHER.decrypt(INPUTSTR.encode('utf-8'))
    return DECRYPTED_STR.decode()

def decrypt_file(INPUTPATH):
    with open(INPUTPATH, 'rb') as INPUTFILE:
        ENCRYPTED_STR = INPUTFILE.read()
    DECRYPTED_STR = decrypt_string(ENCRYPTED_STR)
    return DECRYPTED_STR

def error(msg):
    logging.error(msg)
    if logging.root.level == logging.INFO:
        print(f"ERROR: {msg}")

def info(msg, output=False):
    logging.info(msg)
    if logging.root.level == logging.INFO or output:
        print(f"INFO: {msg}")

def warn(msg):
    logging.warn(msg)
    if logging.root.level == logging.INFO:
        print(f"WARN: {msg}")

def debug(msg):
    logging.debug(msg)
    if logging.root.level == logging.DEBUG:
        print(f"DEBUG: {msg}")
    
def critical(msg):
    logging.critical(msg)
    logging.debug('exception: ', exc_info=True)
    print(f"CRITICAL: {msg}")
    sys.exit()



import os, logging, requests, sys
from socket import gaierror
from json import dumps
from requests.auth import HTTPBasicAuth
    
def post(url=None, headers=None, data=None, json=None, auth=None, expected_codes=[200, 202]):
    return _call("POST", url, headers, data, json, auth, expected_codes)

def patch(url=None, headers=None, data=None, json=None, auth=None, expected_codes=[200]):
    return _call("PATCH", url, headers, data, json, auth, expected_codes)

def get(url=None, headers=None, data=None, json=None, auth=None, expected_codes=[200]):
    return _call("GET", url, headers, data, json, auth, expected_codes)

def put(url=None, headers=None, data=None, json=None, auth=None, expected_codes=[200]):
    return _call("PUT", url, headers, data, json, auth, expected_codes)

def delete(url=None, headers=None, data=None, json=None, auth=None, expected_codes=[200]):
    return _call("DELETE", url, headers, data, json, auth, expected_codes)

def _call(req_type, req_url=None, req_headers=None, req_data=None, req_json=None, req_auth=None, expected_codes=[200]):
    if req_url is None:
        critical('curl2: no URL provided')
        return None
    try:
        if type(req_auth) is dict:
            debug('curl2: adding HTTPBasicAuth to the request')
            RESPONSE = requests.request(method=req_type, url=req_url, headers=req_headers, data=req_data, json=req_json, auth=HTTPBasicAuth(req_auth['user'], req_auth['pass']))
        else:
            RESPONSE = requests.request(method=req_type, url=req_url, headers=req_headers, data=req_data, json=req_json) 
    except gaierror:
        critical(f"curl2: failed to resolve API endpoint - {req_url}")
    if _process_result(RESPONSE, expected_codes):
        return curl_repsonse(RESPONSE)
    else:
        critical(f'unexpected response from the API: {RESPONSE.status_code} \nendpoint: {req_url} \nresponse details:\n {RESPONSE.text}')

def _process_result(result, expected_codes):
    for code in expected_codes:
        if result.status_code == code:
            return True
    return False
    
def _check_arguments(req_url, req_headers, req_data, req_json, req_auth, expected_codes):
    if type(expected_codes) != list and type(req_url) != str and (type(req_headers) == dict or type(req_headers) == None) and (type(req_json) == dict or type(req_json) == None) and (type(req_data) == dict or type(req_data) == None) and (type(req_auth) == dict or type(req_auth) == None):
        return True
    return False

class curl_repsonse():
    def __init__(self, object):
        try:
            self.json=object.json()
        except requests.exceptions.JSONDecodeError:
            self.json=None
        self.status_code=object.status_code
        self.text=object.text
        self.response=object

#### logging

LOGNAME = 'curl2.log'
DICTCONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s %(message)s'
        },
    },
    'handlers': {
        'file_handler': {
            'level': 'INFO',
            'mode': 'a',
            'filename': f"{os.getenv('HOME')}/.{LOGNAME}",
            'class': 'logging.FileHandler',
            'formatter': 'standard'
        }
    },
    'loggers': {
        '': {
            'handlers': ['file_handler'],
            'level': 'INFO',
            'propagate': True
        },
    }
}
logging.config.dictConfig(DICTCONFIG)

def error(msg):
    logging.error(msg)
    if logging.root.level == logging.INFO:
        print(f"ERROR: {msg}")

def info(msg, output=False):
    logging.info(msg)
    if logging.root.level == logging.INFO or output:
        print(f"INFO: {msg}")

def warn(msg):
    logging.warn(msg)
    if logging.root.level == logging.INFO:
        print(f"WARN: {msg}")

def debug(msg):
    logging.debug(msg)
    if logging.root.level == logging.DEBUG:
        print(f"DEBUG: {msg}")
    
def critical(msg):
    logging.critical(msg)
    logging.debug('exception: ', exc_info=True)
    print(f"CRITICAL: {msg}")
    sys.exit(100)
